# Copyright 2017 Big Switch Networks, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslotest import base


class TestImports(base.BaseTestCase):

    # This test simply import files that are not used in other UT.
    # the reasoning for this is that if any upstream dependencies
    # are broken, it is cheapest to find out here.
    def test_imports(self):
        import networking_bigswitch.plugins.bigswitch.bsn_service_plugin  # noqa
        import networking_bigswitch.plugins.bigswitch.extensions.bsnserviceextension  # noqa
